__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from argparse import Namespace
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Generator

from PIL import Image
from imagehash import ImageHash

HashFunction = Callable[[Image.Image], ImageHash]
ImageGroup = Tuple[Path, ...]
ActionFunction = Callable[[Namespace, ImageGroup], Any]
Results = List[ImageGroup]
ResultsGenerator = Generator[List[Path], None, None]
ResultsGrouper = Callable[[ResultsGenerator], Results]
CacheEntry = Tuple[Path, Optional[ImageHash]]
Cache = Dict[Path, ImageHash]



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from functools import wraps
from pathlib import Path
from time import time


def path_with_parent(path: Path) -> str:
    return '/'.join(str(path).split('/')[-2:])


def log_execution_time():
    def actual_decorator(method):
        @wraps(method)
        def allow_fail(self, *args, **kwargs):
            start_time = time()
            return_value = method(self, *args, **kwargs)
            logging.info(
                '%s.%s() run in %.2fs',
                type(self).__name__, method.__name__,
                time() - start_time
            )
            return return_value

        return allow_fail

    return actual_decorator



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from itertools import combinations
from pathlib import Path
from time import time
from typing import Dict, List, Iterator, Optional

from imagehash import ImageHash

from duplicate_images.common import log_execution_time
from duplicate_images.function_types import (
    Cache, HashFunction, ImageGroup, Results, ResultsGenerator, ResultsGrouper
)
from duplicate_images.hash_scanner import ImageHashScanner, ParallelImageHashScanner
from duplicate_images.methods import get_hash_size_kwargs
from duplicate_images.pair_finder_options import PairFinderOptions
from duplicate_images.progress_bar_manager import ProgressBarManager, NullProgressBarManager


def group_results_as_pairs(results: ResultsGenerator) -> Results:
    return [
        pair
        for result in results
        for pair in combinations(list(result), 2)
    ]


def group_results_as_tuples(results: ResultsGenerator) -> Results:
    return [tuple(result) for result in results]


class ImagePairFinder:

    @classmethod
    def create(
            cls, files: List[Path], hash_algorithm: HashFunction,
            options: PairFinderOptions = PairFinderOptions(),
            hash_store: Optional[Cache] = None
    ) -> 'ImagePairFinder':
        hash_size_kwargs = get_hash_size_kwargs(hash_algorithm, options.hash_size)
        group_results = group_results_as_tuples if options.group else group_results_as_pairs
        progress_bars = ProgressBarManager.create(len(files), options.show_progress_bars)
        if not options.parallel:
            scanner = ImageHashScanner(
                files, hash_algorithm, hash_size_kwargs, hash_store, progress_bars
            )
        else:
            scanner = ParallelImageHashScanner(
                files, hash_algorithm, hash_size_kwargs, hash_store, progress_bars
            )
        if options.max_distance == 0 and not options.slow:
            return DictImagePairFinder(
                scanner, group_results, options=options, progress_bars=progress_bars
            )
        if len(files) > 1000:
            logging.warning(
                'Using %s with a big number of images. Expect slow performance.',
                SlowImagePairFinder.__name__
            )
            logging.warning('Consider using [Parallel]DictImagePairFinder instead.')
        return SlowImagePairFinder(scanner, group_results, options, progress_bars)

    def __init__(  # pylint: disable = too-many-arguments
            self, scanner: ImageHashScanner,
            group_results: ResultsGrouper,
            progress_bars: ProgressBarManager = NullProgressBarManager()
    ) -> None:
        self.precalculated_hashes: Dict = {}
        self.group_results = group_results
        self.scanner = scanner
        self.progress_bars = progress_bars
        self.scan_start_time = time()
        logging.info('Using %s', self.__class__.__name__)

    def get_equal_groups(self) -> Results:
        raise NotImplementedError()

    def log_scan_finished(self) -> None:
        logging.info(
            '%d hashes calculated in %.2fs',
            len(self.precalculated_hashes), time() - self.scan_start_time
        )


class DictImagePairFinder(ImagePairFinder):
    """
    Searches by storing the image hashes as keys to a dict.
    Works only if max_distance == 0.
    """
    def __init__(  # pylint: disable = too-many-arguments
            self, scanner: ImageHashScanner,
            group_results: ResultsGrouper,
            options: PairFinderOptions = PairFinderOptions(),
            progress_bars: ProgressBarManager = NullProgressBarManager()
    ) -> None:
        super().__init__(scanner, group_results, progress_bars)
        if options.max_distance != 0:
            raise ValueError('DictImagePairFinder only works if max_distance == 0!')
        self.precalculated_hashes = self.get_hashes()
        self.progress_bars.close_reader()

    @log_execution_time()
    def get_equal_groups(self) -> Results:
        self.progress_bars.close()
        self.log_scan_finished()
        return self.group_results(
            (result for result in self.precalculated_hashes.values() if len(result) > 1)
        )

    def get_hashes(self) -> Dict[ImageHash, List[Path]]:
        hash_dict: Dict[ImageHash, List[Path]] = {}
        for file, image_hash in self.scanner.precalculate_hashes():
            if image_hash is not None:
                hash_dict.setdefault(image_hash, []).append(file)
        return hash_dict


class SlowImagePairFinder(ImagePairFinder):
    """
    Searches by comparing the image hashes of each image to every other, giving O(N^2) performance.
    Does not allow returning the results in groups, only pairs.
    The only option if max_distance != 0.
    """

    def __init__(  # pylint: disable = too-many-arguments
            self, scanner: ImageHashScanner,
            group_results: ResultsGrouper,
            options: PairFinderOptions = PairFinderOptions(),
            progress_bars: ProgressBarManager = NullProgressBarManager()
    ) -> None:
        if group_results is group_results_as_tuples:
            raise ValueError(f'{self.__class__.__name__} only works with pairs, not groups')
        super().__init__(scanner, group_results, progress_bars)
        self.max_distance = options.max_distance or 0
        self.precalculated_hashes = self.get_hashes()
        self.progress_bars.close_reader()

    @log_execution_time()
    def get_equal_groups(self) -> Results:
        self.log_scan_finished()
        image_files = list(self.precalculated_hashes.keys())
        logging.info('Filtering duplicates')
        matches = self.filter_matches(combinations(image_files, 2))
        self.progress_bars.close()
        return matches

    def get_hashes(self) -> Dict[Path, ImageHash]:
        return {
            file: image_hash for file, image_hash in self.scanner.precalculate_hashes()
            if image_hash is not None
        }

    def filter_matches(self, all_pairs: Iterator[ImageGroup]) -> Results:
        self.progress_bars.create_filter_bar(len(self.precalculated_hashes))
        return [
            (file, other_file) for file, other_file in all_pairs
            if self.are_images_equal(file, other_file)
        ]

    def are_images_equal(self, file: Path, other_file: Path) -> bool:
        self.progress_bars.update_filter()
        hash_distance = self.precalculated_hashes[file] - self.precalculated_hashes[other_file]
        logging.debug(
            '%-30s - %-30s = %d', file.stem, other_file.stem, hash_distance
        )
        return hash_distance <= self.max_distance



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging

from multiprocessing.pool import ThreadPool as Pool
from pathlib import Path
from typing import List, Optional, Dict

from PIL import Image

from duplicate_images.common import path_with_parent
from duplicate_images.function_types import Cache, CacheEntry, HashFunction
from duplicate_images.progress_bar_manager import ProgressBarManager, NullProgressBarManager


class ImageHashScanner:

    def __init__(  # pylint: disable = too-many-arguments
            self, files: List[Path], hash_algorithm: HashFunction,
            hash_size_kwargs: Optional[Dict] = None,
            hash_store: Optional[Cache] = None,
            progress_bars: ProgressBarManager = NullProgressBarManager()
    ) -> None:
        self.files = files
        self.algorithm = hash_algorithm
        self.hash_size_kwargs = hash_size_kwargs if hash_size_kwargs is not None else {}
        self.hash_store = hash_store if hash_store is not None else {}
        self.progress_bars = progress_bars

    def precalculate_hashes(self) -> List[CacheEntry]:
        return [self.get_hash(file) for file in self.files]

    def get_hash(self, file: Path) -> CacheEntry:
        self.progress_bars.update_reader()
        try:
            cached = self.hash_store.get(file)
            if cached is not None:
                return file, cached

            image_hash = self.algorithm(Image.open(file), **self.hash_size_kwargs)
            self.hash_store[file] = image_hash
            return file, image_hash
        except OSError as err:
            logging.warning('%s: %s', path_with_parent(file), err)
            return file, None


class ParallelImageHashScanner(ImageHashScanner):

    def precalculate_hashes(self) -> List[CacheEntry]:
        with Pool() as pool:
            return pool.map(self.get_hash, self.files)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from argparse import Namespace
from pathlib import Path
from shlex import quote
from subprocess import call  # nosec
from typing import Any, Callable, Dict, List, Optional

import imagehash

from duplicate_images.common import path_with_parent
from duplicate_images.function_types import ActionFunction, HashFunction, ImageGroup

__all__ = [
    'call', 'quote', 'get_hash_size_kwargs', 'IMAGE_HASH_ALGORITHM', 'ALGORITHM_DEFAULTS',
    'ACTIONS_ON_EQUALITY'
]


def ascending_by_size(group: ImageGroup) -> List[Path]:
    return sorted(group, key=lambda path: (path.stat().st_size, str(path)))


def delete_with_log_message(file: Path) -> None:
    file.unlink()
    logging.info('Deleted %s', path_with_parent(file))


def symlink_to_biggest_file(group: ImageGroup):
    biggest = ascending_by_size(group)[-1]
    others = set(group) - {biggest}
    for file in others:
        delete_with_log_message(file)
        file.symlink_to(biggest)


def shell_exec(args: Namespace, group: ImageGroup) -> None:
    cmd = args.exec
    for num, path in enumerate(group):
        cmd = cmd.replace(f"{'{'}{num+1}{'}'}", f'{quote(str(path))}')
    cmd = cmd.replace('{*}', ' '.join([quote(str(path)) for path in group]))
    call(cmd, shell=True)  # nosec


def get_hash_size_kwargs(algorithm: HashFunction, size: Optional[int]) -> Dict[str, int]:
    if size is None:
        return ALGORITHM_DEFAULTS.get(algorithm, {'hash_size': 8})
    kwarg = next(iter(ALGORITHM_DEFAULTS.get(algorithm, {'hash_size': 8})))
    return {kwarg: size}


IMAGE_HASH_ALGORITHM = {
    'ahash': imagehash.average_hash,
    'phash': imagehash.phash,
    'phash_simple': imagehash.phash_simple,
    'dhash': imagehash.dhash,
    'dhash_vertical': imagehash.dhash_vertical,
    'whash': imagehash.whash,
    'colorhash': imagehash.colorhash,
}  # type: Dict[str, Callable[[Any], imagehash.ImageHash]]

ALGORITHM_DEFAULTS = {
    imagehash.average_hash: {'hash_size': 8},
    imagehash.phash: {'hash_size': 8},
    imagehash.phash_simple: {'hash_size': 8},
    imagehash.dhash: {'hash_size': 8},
    imagehash.dhash_vertical: {'hash_size': 8},
    imagehash.whash: {'hash_size': 8},
    imagehash.colorhash: {'binbits': 3},
}

ACTIONS_ON_EQUALITY: Dict[str, ActionFunction] = {
    'delete-first': lambda args, group: delete_with_log_message(group[0]),
    'd1': lambda args, group: delete_with_log_message(group[0]),
    'delete-last': lambda args, group: delete_with_log_message(group[-1]),
    'dl': lambda args, group: delete_with_log_message(group[-1]),
    'delete-biggest': lambda args, group: delete_with_log_message(ascending_by_size(group)[-1]),
    'd>': lambda args, group: delete_with_log_message(ascending_by_size(group)[-1]),
    'delete-smallest': lambda args, group: delete_with_log_message(ascending_by_size(group)[0]),
    'd<': lambda args, group: delete_with_log_message(ascending_by_size(group)[0]),
    'symlink-smaller': lambda args, group: symlink_to_biggest_file(group),
    'eog': lambda args, group: call(['eog'] + [str(pic) for pic in group]),  # nosec
    'xv': lambda args, group: call(['xv', '-nolim'] + [str(pic) for pic in group]),  # nosec
    'print': lambda args, group: print(*group),
    'print_inline': lambda args, group: print(*group, end=' '),
    'quote': lambda args, group: print(' '.join([quote(str(pic)) for pic in group])),
    'quote_inline': lambda args, group: print(
        ' '.join([quote(str(pic)) for pic in group]), end=' '
    ),
    'exec': lambda args, group: shell_exec(args, group),  # pylint: disable=unnecessary-lambda
    'none': lambda args, group: None
}



__author__ = 'lene'

from argparse import ArgumentParser, Namespace
from typing import List, Optional

from duplicate_images.methods import ACTIONS_ON_EQUALITY, IMAGE_HASH_ALGORITHM


def parse_command_line(args: Optional[List[str]] = None) -> Namespace:
    parser = ArgumentParser(description='Find pairs of equal or similar images.')

    parser.add_argument(
        'root_directory', default='.', nargs='+',
        help='The root of the directory tree under which images are compared'
    )
    parser.add_argument(
        '--algorithm', choices=IMAGE_HASH_ALGORITHM.keys(),
        default='phash',
        help='Method used to determine if two images are considered equal'
    )
    parser.add_argument(
        '--max-distance', type=int, default=0,
        help='Maximum hash distance for images to be considered equal'
    )
    parser.add_argument(
        '--hash-size', type=int,
        help='Hash size (or number of bin bits for colorhash)'
    )
    parser.add_argument(
        '--on-equal', choices=ACTIONS_ON_EQUALITY.keys(),
        default='print', help='Command to be run on each pair of images found to be equal'
    )
    parser.add_argument(
        '--exec', type=str,
        help='Command to execute (replaces {1}, {2} with file pathes)'
    )
    parser.add_argument(
        '--parallel', action='store_true', help='Calculate hashes using all available cores'
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--slow', action='store_true', help='Use slow (O(N^2)) algorithm'
    )
    group.add_argument(
        '--group', action='store_true',
        help='Handle equal images in a group instead of multiple pairs'
    )
    parser.add_argument(
        '--progress', action='store_true', help='Show progress bars during processing'
    )
    parser.add_argument(
        '--debug', action='store_true', help='Print lots of debugging info'
    )
    parser.add_argument(
        '--quiet', '-q', action='count', default=0, help='Decrease log level by one for each'
    )
    parser.add_argument(
        '--hash-db', default=None, help='File storing precomputed hashes'
    )

    namespace = parser.parse_args(args)
    if namespace.on_equal == 'exec' and not namespace.exec:
        parser.error('--exec argument is required')
    return namespace



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import json
import logging
import pickle  # nosec
from pathlib import Path
from typing import Any, IO, Callable, Optional, Union, Dict, Tuple

from imagehash import ImageHash, hex_to_hash

from duplicate_images.common import log_execution_time
from duplicate_images.function_types import Cache


class NullHashStore:

    def __init__(self) -> None:
        logging.info('No persistent storage for calculated image hashes set up')

    def __enter__(self) -> Cache:
        return {}

    def __exit__(self, _: Any, __: Any, ___: Any) -> None:
        pass


class FileHashStore:

    @classmethod
    def create(
            cls, store_path: Optional[Path], algorithm: str, hash_size_kwargs: Dict
    ) -> Union['FileHashStore', NullHashStore]:
        if store_path is None:
            return NullHashStore()
        if store_path.suffix == '.pickle':
            return PickleHashStore(store_path, algorithm, hash_size_kwargs)
        return JSONHashStore(store_path, algorithm, hash_size_kwargs)

    def __init__(self, store_path: Path, algorithm: str, hash_size_kwargs: Dict) -> None:
        self.store_path = store_path
        self.algorithm = algorithm
        self.hash_size_kwargs = hash_size_kwargs
        self.values: Cache = {}
        try:
            self.load()
            logging.info(
                'Opened persistent storage %s with %d entries', store_path, len(self.values)
            )
        except (FileNotFoundError, EOFError, pickle.PickleError):
            logging.info('Creating new persistent storage at %s', store_path)

    def __enter__(self) -> Cache:
        return self.values

    def __exit__(self, _: Any, __: Any, ___: Any) -> None:
        if self.store_path.is_file():
            if self.store_path.with_suffix('.bak').is_file():
                self.store_path.with_suffix('.bak').unlink()
            self.store_path.rename(self.store_path.with_suffix('.bak'))
        self.dump()

    def metadata(self) -> Dict:
        return {'algorithm': self.algorithm, **self.hash_size_kwargs}

    def values_with_metadata(self) -> Tuple[Dict, Dict]:
        return self.values, self.metadata()

    def checked_load(self, file: IO, load: Callable[[IO], Tuple[Cache, Dict]]) -> None:
        try:
            values, metadata = load(file)  # nosec
        except IndexError as error:
            raise ValueError('Save file not in format: [values, metadata]') from error
        if not isinstance(values, dict):
            raise ValueError(f'Not a dict: {values}')
        if not metadata:
            raise ValueError('Metadata empty')
        if not isinstance(metadata, dict):
            raise ValueError(f'Metadata not a dict: {metadata}')
        bad_keys = [key for key in values.keys() if not isinstance(key, Path)]
        if bad_keys:
            raise ValueError(f'Not a Path: {bad_keys}')
        bad_values = [value for value in values.values() if not isinstance(value, ImageHash)]
        if bad_values:
            raise ValueError(f'Not an image hash: {bad_values}')
        if metadata['algorithm'] != self.algorithm:
            raise ValueError(f'Algorithm mismatch: {metadata["algorithm"]} != {self.algorithm}')
        if metadata.keys() != self.metadata().keys():
            raise ValueError(f'Metadata mismatch: {metadata} != {self.metadata()}')
        if metadata != self.metadata():
            raise ValueError(f'Metadata mismatch: {metadata} != {self.metadata()}')
        self.values = values

    def load(self) -> None:
        raise NotImplementedError()

    def dump(self) -> None:
        raise NotImplementedError()


class PickleHashStore(FileHashStore):

    @log_execution_time()
    def load(self) -> None:
        with self.store_path.open('rb') as file:
            self.checked_load(file, pickle.load)

    @log_execution_time()
    def dump(self) -> None:
        with self.store_path.open('wb') as file:
            pickle.dump(self.values_with_metadata(), file)  # nosec


def load_values_and_metadata(file: IO) -> Tuple[Cache, Dict]:
    try:
        valds = json.load(file)
    except json.JSONDecodeError as error:
        raise ValueError('Save file not in JSON format') from error
    if not isinstance(valds, list):
        raise ValueError('Save file not in format: [values, metadata]')
    if not isinstance(valds[0], dict):
        raise ValueError(f'Not a dict: {valds[0]}')
    if not isinstance(valds[1], dict):
        raise ValueError(f'Metadata not a dict: {valds[1]}')
    return {Path(k): hex_to_hash(str(v)) for k, v in valds[0].items()}, valds[1]


class JSONHashStore(FileHashStore):

    @log_execution_time()
    def load(self) -> None:
        with self.store_path.open('r') as file:
            self.checked_load(file, load_values_and_metadata)

    # see https://bugs.python.org/issue18820 for why this pain is necessary (Python does not allow
    # to automatically convert dict keys for JSON export
    def converted_values(self):
        return {str(k.resolve()): str(v) for k, v in self.values.items()}

    @log_execution_time()
    def dump(self) -> None:
        with self.store_path.open('w') as file:
            json.dump((self.converted_values(), self.metadata()), file)


HashStore = Union[NullHashStore, PickleHashStore, JSONHashStore]
